/*	AVS coroutine module for controlling the input of raw data from a remote
 *	supercomputer via internet port to the named local pipe /tmp/magicfile,
 * 	where an AVS read_whatever module can find it. When a remote connection
 *  is established and the flow executive has completed computation, the 
 *  module sends a "ready to receive" signal to the other process and outputs
 *  the filename of the FIFO on it's outport port. The module and read_whatever
 *  then hang until data is sent.
 * 	8/3/93 by Bevan C Bennett
 *	No stairway to heaven.
 */

#include <stdio.h>
#include <avs/avs.h>
#include <avs/flow.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define RDY	"R"
#define HUP	"H"
#define SOCK	6669
#define Mdefault	"/tmp/charon"
#define MODE	0010666
#define CS	"Connect to Host"
#define DS	"Disconnect from Host"

static int connected = 0;
static int situation = 1;
static int paused = 1;
static int s, fd;
static struct sockaddr_in sin;
static struct hostent *hp;
static char *MFile;

/* AVS initialization procedure, called by AVS when module is loaded. */

data_remote()
{
int p;
int data_remote_exit();

	AVSset_module_name("Charon", MODULE_DATA);
	AVScreate_output_port("Poke","string");

	p = AVSadd_parameter("Remote Host", "string", 0, 0, 0);
	AVSadd_parameter_prop(p, "width", "integer", 4);		
	
	p = AVSadd_parameter("Remote Port", "integer", SOCK, 1025, 32767);
	AVSconnect_widget(p,"typein_integer");
	AVSadd_parameter_prop(p, "width", "integer", 4);
	
	p = AVSadd_parameter("Local Pipe","string",Mdefault,0,0);
	AVSconnect_widget(p, "none");
	AVSadd_parameter_prop(p, "width", "integer", 4);

	p = AVSadd_parameter(CS, "boolean", 0, 0, 1);
	AVSadd_parameter_prop(p, "width", "integer", 4);

	p = AVSadd_parameter("Pause","boolean", 1, 0, 1);
	AVSadd_parameter_prop(p, "width","integer",4);
	AVSadd_parameter_prop(p, "title","string","                    ++Pause++");

	AVSset_destroy_proc(data_remote_exit);
}


main(argc,argv)

int argc;
char *argv[];
{
int  data_remote();
char *remhost;
char *avspipe;
int  remport;
int  remhup;
int  done,j,a,b;
int	 i;
unsigned char n[8];
char buf[BUFSIZ], line[1];

	MFile=Mdefault;
	AVScorout_init(argc,argv,data_remote);
	AVScorout_set_sync(0);
	unlink(MFile);
	if ((mknod(MFile, MODE, 0)) < 0) {
		AVSfatal("%s\n%s\n","Failed to create FIFO.","Module Exiting.");
		exit(1);
	}
	while (1) {
	
		if ((connected) && (!paused)) AVScorout_mark_changed();
		AVScorout_wait();
		AVScorout_input(&remhost,&remport,&avspipe,&remhup,&paused);

		if (AVSparameter_changed("Local Pipe") && strcmp(MFile,avspipe)) {
			if (paused) {
				unlink(avspipe);
				if ((mknod(avspipe, MODE, 0)) < 0) {
					AVSwarning("Couldn't create FIFO %s.\n",avspipe);	
					AVSmodify_parameter("Local Pipe",AVS_VALUE,MFile,0,0);
				} else {
					unlink(MFile);
					MFile=avspipe;
				}
			} else {
				AVSmodify_parameter("Local Pipe",AVS_VALUE,MFile,0,0);
			}
		}

		if (remhup) {
			AVSmodify_parameter(CS,AVS_VALUE,0,0,0);
			remhup=0;
			if (!connected) {
				AVSmodule_status("Looking for host...",11);
				if ((hp = gethostbyname(remhost)) == NULL) {
					AVSwarning("%s: Unknown Host\n",remhost);
					AVSmodify_parameter("Remote Host", AVS_VALUE, 0, 0, 0);
					situation = 0;
				}
				else if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
					AVSwarning("%s\n","Failed to open socket.");
					situation = 0;	
				}
				else situation = 1;
				if (situation) {
					sin.sin_family = AF_INET;
					sin.sin_port = htons(remport);
					bcopy(hp->h_addr, &sin.sin_addr, hp->h_length);
					if (connect(s, &sin, sizeof(sin)) < 0) {
						AVSwarning("%s %s, port %d.\n",
							   "Connect failed to server",remhost, remport);
						situation = 0;
					}
					else situation = 1;
				}
				connected = situation;
				if (connected) 
					AVSmodify_parameter_prop(CS,"title","string",DS);
			}
			else {
				hang_up();
			}
		}
		
		if ((connected) && (!paused)) {
			AVSmodule_status("Waiting for data.",10);
			AVScorout_exec();
			send(s,RDY,sizeof(RDY),0);
			recv(s,line,sizeof(RDY),0);
			AVScorout_output(MFile);
			while ((fd=open(MFile,O_WRONLY|O_NDELAY)) < 0) 
				sleep(1);
			close(fd);
			fd = open(MFile,O_WRONLY);
			done=0;
			while (!done) {
				i = recv(s,n,sizeof(n),0);
				i = n[0]+(n[1]<<8)+(n[2]<<16)+(n[3]<<24);
				j=0; a=i;
				if (a>BUFSIZ) b=BUFSIZ; else b=a;
				while (a>0) {
					j = recv(s,buf,b,0);
					write(fd,buf,j);
					a = a-j;
					if (a>BUFSIZ) b=BUFSIZ; else b=a;	
				}
				if (n[7]&2) hang_up();
				if (n[7]&1) done=1;
			}
			close(fd);
		}
	}	/* while */
}

data_remote_exit()
{
	close(fd);
	unlink(MFile);
	hang_up();
}

hang_up()
{
	if (connected) {
		AVSmodify_parameter_prop(CS,"title","string",CS);
		send(s,HUP,sizeof(HUP),0);
		shutdown(s,2);
	}
	connected = 0;
	paused = 1;
	AVSmodify_parameter("Pause",AVS_VALUE,1);
}
